"""
    AwesomeTkinter, a new tkinter widgets design using custom styles and images

    :copyright: (c) 2020-2021 by Mahmoud Elshahat.

"""

import tkinter as tk
from tkinter import ttk
from .utils import *
from .config import *
from .images import *
from .scrollbar import SimpleScrollbar


class ScrollableFrame(tk.Frame):
    """A frame with scrollbars
    inspired by : https://stackoverflow.com/a/3092341
    basically it is a frame inside a canvas inside another frame

    usage:
        frame = ScrollableFrame(root)
        frame.pack(fill='both', expand=True)

        # add your widgets normally
        tk.Label(frame, text=hello).pack()
    """

    def __init__(self, parent, vscroll=True, hscroll=True, autoscroll=False, bg=None, sbar_fg=None, sbar_bg=None,
                 vbar_width=10, hbar_width=10):
        """initialize

        Args:
            parent (tk.Widget): tkinter master widget
            vscroll (bool): use vertical scrollbar
            hscroll (bool): use horizontal scrollbar
            autoscroll (bool): auto scroll to bottom if new items added to frame
            bg (str): background
            sbar_fg (str): color of scrollbars' slider
            sbar_bg (str): color of scrollbars' trough, default to frame's background
            vbar_width (int): vertical scrollbar width
            hbar_width (int): horizontal scrollbar width
        """
        self.autoscroll = autoscroll
        self.current_height = None

        sbar_bg = sbar_bg or 'white'
        sbar_fg = sbar_fg or 'blue'

        # create outside frame
        self.outer_frame = tk.Frame(parent, bg=bg)

        # create canvas
        self.canvas = tk.Canvas(self.outer_frame, borderwidth=0, highlightthickness=0, background=bg)

        # initialize super class
        tk.Frame.__init__(self, self.canvas, bg=bg)

        # scrollbars
        if vscroll:
            self.vsb = SimpleScrollbar(self.outer_frame, orient="vertical", command=self.canvas.yview, bg=sbar_bg,
                                       slider_color=sbar_fg, width=vbar_width)
            self.canvas.configure(yscrollcommand=self.vsb.set)

            self.vsb.pack(side="right", fill="y")
        if hscroll:
            self.hsb = SimpleScrollbar(self.outer_frame, orient="horizontal", command=self.canvas.xview, bg=sbar_bg,
                                       slider_color=sbar_fg, width=hbar_width)
            self.canvas.configure(xscrollcommand=self.hsb.set)

            self.hsb.pack(side="bottom", fill="x")

        self.canvas.pack(side="left", fill="both", expand=True)

        self._id = self.canvas.create_window((0, 0), window=self, anchor="nw", tags="self")

        self.bind("<Configure>", self._on_self_configure)
        self.canvas.bind("<Configure>", self._on_canvas_configure)

        # scroll with mousewheel
        scroll_with_mousewheel(self, target=self.canvas)

        # use outer frame geometry managers
        self.pack = self.outer_frame.pack
        self.pack_forget = self.outer_frame.pack_forget

        self.grid = self.outer_frame.grid
        self.grid_forget = self.outer_frame.grid_forget
        self.grid_remove = self.outer_frame.grid_remove

        self.place = self.outer_frame.place
        self.place_forget = self.outer_frame.place_forget

        # get scroll methods from canvas
        self.yview_scroll = self.canvas.yview_scroll
        self.xview_scroll = self.canvas.xview_scroll
        self.yview_moveto = self.canvas.yview_moveto
        self.xview_moveto = self.canvas.xview_moveto

    def _on_self_configure(self, event):
        """Reset the scroll region to match contents"""
        if self.winfo_height() != self.current_height:
            self.canvas.configure(scrollregion=self.canvas.bbox("all"))

            # scroll to bottom, if new widgets added to frame
            if self.autoscroll:
                self.scrolltobottom()

            self.current_height = self.winfo_height()

    def _on_canvas_configure(self, event):
        """expand self to fill canvas"""
        self.canvas.itemconfigure(self._id, width=self.canvas.winfo_width())

    def vscroll(self, fraction):
        """scroll canvas vertically

        Args:
            fraction (float): from 0 "top" to 1.0 "bottom"
        """

        self.canvas.yview_moveto(fraction)

    def scrolltobottom(self):
        self.vscroll(1.0)

    def scrolltotop(self):
        self.vscroll(0)

    def hscroll(self, fraction):
        """scroll canvas horizontally

        Args:
            fraction (float): from 0 "left" to 1.0 "right"
        """

        self.canvas.xview_moveto(fraction)


class Frame3d(ttk.Frame):
    """create a frame with 3d background color and shadow"""
    styles = []

    def __init__(self, parent, bg=None, **options):
        """initialize

        Args:
            parent: tkinter container widget, i.e. root or another frame
            bg (str): color of frame
        """
        self.bg = bg or DEFAULT_COLOR
        parent_color = get_widget_attribute(parent, 'background') or DEFAULT_COLOR

        # initialize super class
        ttk.Frame.__init__(self, parent, **options)

        # create unique style name based on frame color
        frame_style = f'Frame3d_{generate_unique_name(color_to_rgba(self.bg))}'

        # create style
        if frame_style not in Frame3d.styles:

            self.img = self.create_image()

            # create elements
            s = ttk.Style()
            element_style = f'{frame_style}_element'
            s.element_create(element_style, 'image', self.img, border=15, sticky="nsew")
            s.layout(frame_style, [(element_style, {"sticky": "nsew"})])
            s.map(frame_style, background=[('', parent_color)])

            # add to styles
            Frame3d.styles.append(frame_style)

        self['style'] = frame_style

    def create_image(self):

        shadow_img = create_pil_image(b64=btn_base)
        img = create_pil_image(b64=btn_face, color=self.bg)

        # merge face with base image
        img = mix_images(shadow_img, img)

        return ImageTk.PhotoImage(img) 



"""
    AwesomeTkinter, a new tkinter widgets design using custom styles and images

    :copyright: (c) 2020-2021 by Mahmoud Elshahat.

"""

import tkinter as tk
from tkinter import ttk
from .utils import *
from .config import *
from .images import *


class Button3d(ttk.Button):
    """create a button with 3d background color and shadow"""
    styles = []

    def __init__(self, parent, bg=None, fg=None, **options):
        """initialize

        Args:
            parent: tkinter container widget, i.e. root or another frame
            bg (str): button color
            fg (str): text color
        """
        self.bg = bg or DEFAULT_COLOR
        parent_color = get_widget_attribute(parent, 'background') or DEFAULT_COLOR

        # initialize super class
        ttk.Button.__init__(self, parent, **options)

        # create unique style name based on frame color
        button_style = f'Button3d_{len(Button3d.styles)}_.TButton'

        # create style
        if button_style not in Button3d.styles:
            self.create_images()

            # create elements
            s = ttk.Style()
            element_name = f'Button.focus.{button_style}_element'
            s.element_create(element_name, 'image', self.img, ('pressed', self.pressed_img), border=12, sticky="nsew")

            s.layout(button_style,
                     [(element_name, {'sticky': 'nswe', 'children':
                         [('Button.padding', {'sticky': 'nswe', 'children':
                             [('Button.label', {'sticky': 'nswe'})]})]})])

            s.map(button_style, background=[('', parent_color)], foreground=[('', calc_font_color(self.bg))])
            s.configure(button_style, padding=0, borderwidth=0, focuscolor=self.bg)

            # add to styles
            Button3d.styles.append(button_style)

        self['style'] = button_style

    def create_images(self):

        shadow_img = create_pil_image(b64=btn_base)
        img = create_pil_image(b64=btn_face, color=self.bg)

        # merge face with base image
        img = mix_images(shadow_img, img)

        pressed_img = img.rotate(180)

        self.img = ImageTk.PhotoImage(img)
        self.pressed_img = ImageTk.PhotoImage(pressed_img) 


class Radiobutton(ttk.Radiobutton):
    """ttk.Radiobutton with better indicator quality"""

    styles = []
    imgs = {'outline_img': {}, 'selection_img': {}}

    def __init__(self, parent, text=None, ind_bg=None, ind_mark_color=None, ind_outline_color=None, bg=None, fg=None,
                 font=None, value=None, **kwargs):
        """initialize
        Args:
            parent: tkinter container
            text (str): text
            ind_bg (str): indicator ring background "fill color"
            ind_outline_color (str): indicator outline / ring color
            ind_mark_color (str): check mark color
            bg (str): background color "should match parent bg
            fg (str): text color
            font (str): text font, e.g. "any 10 bold"
            value (any): value assigned to button variable when selected
        """
        bg = bg or get_widget_attribute(parent, 'background')
        fg = fg or calc_font_color(bg)
        ind_bg = ind_bg or bg
        ind_outline_color = ind_outline_color or fg
        ind_mark_color = ind_mark_color or fg

        value = value if value is not None else text

        custom_style = f'RadioButton_{len(Radiobutton.styles)}'
        Radiobutton.styles.append(custom_style)

        s = ttk.Style()

        self.outline_img = Radiobutton.imgs['outline_img'].get(ind_outline_color)
        self.selection_img = Radiobutton.imgs['selection_img'].get(ind_mark_color)

        # create tkinter PhotoImage
        if not (self.outline_img and self.selection_img):
            # create indicator outline
            outline_img = create_circle(size=12, thickness=1, color=ind_outline_color, fill=ind_bg, offset=2)

            # create indicator mark
            mark_img = create_circle(size=6, thickness=0, color=ind_mark_color, fill=ind_mark_color)

            selection_img = mix_images(outline_img, mark_img)

            # create tkinter PhotoImage
            self.outline_img = create_image(img=outline_img)
            self.selection_img = create_image(img=selection_img)

            # store them for future use
            Radiobutton.imgs['outline_img'][ind_outline_color] = self.outline_img
            Radiobutton.imgs['selection_img'][ind_mark_color] = self.selection_img

        # custom style
        s.layout(custom_style, [('Radiobutton.padding',
                                 {'sticky': 'nswe', 'children': [('Radiobutton.label', {'sticky': 'nswe'})]})])
        s.configure(custom_style, foreground=fg)
        s.map(custom_style, background=[('', bg)])

        if font:
            s.configure(custom_style, font=font)

        # set default options
        options = dict(
            text=text,
            value=value,
            style=custom_style,
            image=[self.outline_img, 'selected', self.selection_img],
            compound='left'
        )

        # override default options
        options.update(**kwargs)

        # initialize super class
        ttk.Radiobutton.__init__(self, master=parent, **options)


class Checkbutton(tk.Checkbutton):
    """ tk.checkbutton with images """
    imgs = {'empty_box_img': {}, 'checked_box_img': {}}

    def __init__(self, parent, box_color=None, check_mark_color=None, text_color=None, size=None, **kwargs):
        """initialize
        Args:
            parent: tkinter container
            box_color (str): checkbox outline color
            check_mark_color (str): check mark color
        """
        bg = kwargs.get('bg') or get_widget_attribute(parent, 'background') # bg of checkbox
        check_mark_color = check_mark_color or calc_font_color(bg)
        box_color = box_color or calc_font_color(bg)
        text_color = text_color or calc_font_color(bg)

        # images
        self.empty_box_img = Checkbutton.imgs['empty_box_img'].get(box_color)
        self.checked_box_img = Checkbutton.imgs['checked_box_img'].get(check_mark_color)

        # create tkinter PhotoImage
        if not (self.empty_box_img and self.checked_box_img):
            self.empty_box_img = create_image(b64=unchecked_icon, color=box_color, size=size)
            self.checked_box_img = create_image(b64=checked_icon, color=check_mark_color, size=size)

            # store images for future use
            Checkbutton.imgs['empty_box_img'][box_color] = self.empty_box_img
            Checkbutton.imgs['checked_box_img'][check_mark_color] = self.checked_box_img

        # set default options
        options = dict(
            bg=bg,
            activebackground=bg,
            highlightbackground=bg,

            fg=text_color,
            activeforeground=text_color,
            selectcolor=bg,

            compound='left',

            indicatoron=False,  # disable original indicator
            image=self.empty_box_img,
            selectimage=self.checked_box_img
        )

        # add space before text
        text = kwargs.get('text')
        if text:
            kwargs['text'] = ' ' + text

        # override default options
        options.update(**kwargs)

        # initialize super class
        tk.Checkbutton.__init__(self, master=parent, **options)


__all__ = ['Button3d', 'Radiobutton', 'Checkbutton']



"""
    AwesomeTkinter, a new tkinter widgets design using custom styles and images

    :copyright: (c) 2020 by Mahmoud Elshahat.

"""

import tkinter as tk
from tkinter import ttk
from .utils import *
from .scrollbar import SimpleScrollbar


class ScrolledText(tk.Frame):
    """Scrolled multiline entry good for log output

    has both horizontal and vertical scrollbar

    auto-scroll vertically by default, if you move vertical scrollbar it will stop auto  scroll until vertical
    scroll bar moved back to bottom

    undo action disabled to save memory

    you can pass extra parameter to Text thru ScrolledText().text.config(**options)
    """

    def __init__(self, parent, bg='white', fg='black', bd=0, wrap=None, vscroll=True, hscroll=True, autoscroll=True,
                 max_chars=None, sbar_fg=None, sbar_bg=None, vbar_width=10, hbar_width=10, **kwargs):
        """initialize

        Args:
            parent (tk.Frame): parent widget
            bg (str): background color
            fg (str): foreground color
            bd (int): border width
            wrap (bool): wrap text, if omitted it will be true if no hscroll
            vscroll (bool): include vertical scrollbar
            hscroll (bool): include horizontal scrollbar
            autoscroll (bool): automatic vertical scrolling
            max_chars (int): maximum characters allowed in Text widget, text will be truncated from the beginning to
                             match the max chars
            sbar_fg (str): color of scrollbars' slider
            sbar_bg (str): color of scrollbars' trough, default to frame's background
            vbar_width (int): vertical scrollbar width
            hbar_width (int): horizontal scrollbar width

        """
        tk.Frame.__init__(self, parent, bg=bg)

        self.bd = bd
        self.bg = bg
        self.fg = fg
        self.vscroll = vscroll
        self.hscroll = hscroll
        self.autoscroll = autoscroll
        self.max_chars = max_chars
        self.sbar_bg = sbar_bg
        self.sbar_fg = sbar_fg

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.var = tk.StringVar()

        # wrap mechanism
        if wrap or not hscroll:
            wrap = tk.WORD
        else:
            wrap = 'none'

        self.text = tk.Text(self, bg=self.bg, fg=self.fg, bd=self.bd, wrap=wrap, undo='false', **kwargs)
        self.text.grid(sticky='ewns')

        if self.vscroll:
            self.vbar = SimpleScrollbar(self, orient='vertical', command=self.text.yview, slider_color=self.sbar_fg,
                                        bg=self.sbar_bg, width=vbar_width)
            self.vbar.grid(row=0, column=1, sticky='ns')
            self.text.config(yscrollcommand=self.vbar.set)

        if self.hscroll:
            self.hbar = SimpleScrollbar(self, orient='horizontal', command=self.text.xview, slider_color=self.sbar_fg,
                                        bg=self.sbar_bg, width=hbar_width)
            self.hbar.grid(row=1, column=0, sticky='ew')
            self.text.config(xscrollcommand=self.hbar.set)

        # bind mouse wheel to scroll
        scroll_with_mousewheel(self.text)

        # aliases
        self.delete = self.text.delete
        self.get = self.text.get
        self.delete = self.text.delete
        self.insert = self.text.insert

    def set(self, text):
        """replace contents"""
        self.clear()

        if self.max_chars:
            count = len(text)
            if count > self.max_chars:
                delta = count - self.max_chars
                text = text[delta:]

        self.text.insert("1.0", text)

        self.scrolltobottom()

    def clear(self):
        """clear all Text widget contents"""
        self.text.delete("1.0", tk.END)

    def append(self, text, text_color=None, text_bg=None):
        """append text with arbitrary colors"""

        color_tags = []

        if text_color:
            self.text.tag_configure(text_color, foreground=text_color)
            color_tags.append(text_color)

        if text_bg:
            self.text.tag_configure(text_bg, foreground=text_bg)
            color_tags.append(text_bg)

        self.text.insert(tk.END, text, ','.join(color_tags))

        self.remove_extra_chars()

        self.scrolltobottom()

    def remove_extra_chars(self):
        """remove characters from beginning of Text widget if it exceeds max chars"""
        if self.max_chars:
            # get current text characters count
            # txt = self.text.get()
            count = len(self.text.get("1.0", tk.END))
            if count > self.max_chars:
                delta = count - self.max_chars
                self.text.delete("1.0", f"1.0 + {delta} chars")

    def scrolltobottom(self):
        """scroll to bottom if autoscroll enabled and scrollbar position at the bottom"""
        try:
            if self.autoscroll and self.vbar.get()[1] == 1:
                self.text.yview_moveto("1.0")
        except:
            pass 



"""
    AwesomeTkinter, a new tkinter widgets design using custom styles and images

    :copyright: (c) 2020-2021 by Mahmoud Elshahat.

"""

import tkinter as tk


class RightClickMenu(tk.Menu):
    """Context menu or right click menu popup"""
    def __init__(self, parent, menu_items, callback=None, bg='white', fg='black', abg='blue', afg='white'):
        """initialize
        Args:
            parent: tkinter widget to show this menu when right clicked
            menu_items (iterable): a list of entries / options to show in right click menu, to add a separator you can
                                   pass a string '---' as an item name
            callback (callable): any function or method that will be called when an option get selected,
                                 should expect option name as a positional argument
            bg: background color
            fg: text color
            abg: background color on mouse hover
            afg: text color on mouse hover

        Example:
            right_click_map = {'say hi': lambda x: print(x),  # x = 'say hi'
                               'help': show_help,
                               'blah blah': blah_callback,
                               }

            RightClickMenu(my_listbox, right_click_map.keys(),
                           callback=lambda option: right_click_map[option](),
                           bg='white', fg='black', abg=blue, afg='white')
        """
        self.callback = callback

        # initialize super
        tk.Menu.__init__(self, parent, tearoff=0, bg=bg, fg=fg, activebackground=abg, activeforeground=afg)

        for option in menu_items:
            if option == '---':
                self.add_separator()
            else:
                self.add_command(label=f' {option}', command=lambda x=option: self.context_menu_handler(x))

        parent.bind("<Button-3>", self.popup, add='+')
        parent.bind("<Button-2>", self.popup, add='+')

        self.parent = parent

    def popup(self, event):
        """show right click menu"""
        x, y = event.x_root, event.y_root
        self.tk_popup(x, y)

    def context_menu_handler(self, option):
        """handle selected option

        Args:
            option (str): selected option
        """

        if callable(self.callback):
            self.callback(option)



#!/usr/bin/env python
"""
    AwesomeTkinter, a new tkinter widgets design using custom styles and images

    :copyright: (c) 2020 by Mahmoud Elshahat.

"""

import tkinter as tk
from tkinter import ttk

from .button import Button3d, Radiobutton, Checkbutton
from .frame import Frame3d, ScrollableFrame
from .menu import RightClickMenu
from .progressbar import RadialProgressbar, RadialProgressbar3d
from .scrollbar import SimpleScrollbar
from .text import ScrolledText
from .utils import *
from .config import *
from .version import __version__


def main():
    
    root = tk.Tk()
    root.title('AwesomeTkinter ' + __version__)
    root.config(background=DEFAULT_COLOR)

    # select tkinter theme required for things to be right on windows,
    # only 'alt', 'default', or 'classic' can work fine on windows 10
    s = ttk.Style()
    s.theme_use('default')

    f1 = Frame3d(root)
    f1.pack(side='left', expand=True, fill='both', padx=3, pady=3)

    bar = RadialProgressbar3d(f1, fg='cyan', size=120, font_size_ratio=0.08)
    bar.pack(padx=20, pady=20)
    bar.start()

    Button3d(f1, text='3D Button').pack(pady=10)

    f2 = Frame3d(root)
    f2.pack(side='left', expand=True, fill='both', padx=3, pady=3)

    bar = RadialProgressbar(f2, fg='green')
    bar.pack(padx=30, pady=30)
    bar.start()

    btn = Button3d(f2, text='Pressed Button')
    btn.pack(pady=10)

    root.mainloop()















 



"""
    AwesomeTkinter, a new tkinter widgets design using custom styles and images

    :copyright: (c) 2020-2021 by Mahmoud Elshahat.

"""

import tkinter as tk
from tkinter import ttk
from .utils import *
from .images import *


class RadialProgressbar(tk.Frame):
    """create radial flat progressbar

    basically this is a ttk horizontal progressbar modified using custom style layout and images

    Example:
        bar = RadialProgressbar(frame1, size=150, fg='green')
        bar.grid(padx=10, pady=10)
        bar.start()
    """

    # class variables to be shared between objects
    styles = []  # hold all style names created for all objects
    imgs = {}  # imgs{"size":{"color": img}}  example: imgs{"100":{"red": img}}

    def __init__(self, parent, size=100, bg=None, fg='cyan', text_fg=None, text_bg=None, font=None, font_size_ratio=0.1,
                 base_img=None, indicator_img=None, parent_bg=None, **extra):
        """initialize progressbar

        Args:
            parent  (tkinter object): tkinter container, i.e. toplevel window or frame
            size (int or 2-tuple(int, int)) size of progressbar in pixels
            bg (str): color of base ring
            fg(str): color of indicator ring
            text_fg (str): percentage text color
            font (str): tkinter font for percentage text, e.g. 'any 20'
            font_size_ratio (float): font size to progressbar width ratio, e.g. for a progressbar size 100 pixels,
                                     a 0.1 ratio means font size 10
            base_img (tk.PhotoImage): base image for progressbar
            indicator_img (tk.PhotoImage): indicator image for progressbar
            parent_bg (str): color of parent container
            extra: any extra kwargs

        """

        self.parent = parent
        self.parent_bg = parent_bg or get_widget_attribute(self.parent, 'background')
        self.bg = bg or calc_contrast_color(self.parent_bg, 30)
        self.fg = fg
        self.text_fg = text_fg or calc_font_color(self.parent_bg)
        self.text_bg = text_bg or self.parent_bg
        self.size = size if isinstance(size, (list, tuple)) else (size, size)
        self.font_size_ratio = font_size_ratio
        self.font = font or f'any {int((sum(self.size) // 2) * self.font_size_ratio)}'

        self.base_img = base_img
        self.indicator_img = indicator_img

        self.var = tk.IntVar()

        # initialize super class
        tk.Frame.__init__(self, master=parent)

        # create custom progressbar style
        self.bar_style = self.create_style()

        # create tk Progressbar
        self.bar = ttk.Progressbar(self, orient='horizontal', mode='determinate', length=self.size[0],
                                   variable=self.var, style=self.bar_style)
        self.bar.pack()

        # percentage Label
        self.percent_label = ttk.Label(self.bar, text='0%')
        self.percent_label.place(relx=0.5, rely=0.5, anchor="center")

        # trace progressbar value to show in label
        self.var.trace_add('write', self.show_percentage)

        # set default attributes
        self.config(**extra)

        self.start = self.bar.start
        self.stop = self.bar.stop

    def set(self, value):
        """set and validate progressbar value"""
        value = self.validate_value(value)
        self.var.set(value)

    def get(self):
        """get validated progressbar value"""
        value = self.var.get()
        return self.validate_value(value)

    def validate_value(self, value):
        """validate progressbar value
        """

        try:
            value = int(value)
            if value < 0:
                value = 0
            elif value > 100:
                value = 100
        except:
            value = 0

        return value

    def create_style(self):
        """create ttk style for progressbar

        style name is unique and will be stored in class variable "styles"
        """

        # create unique style name
        bar_style = f'radial_progressbar_{len(RadialProgressbar.styles)}'

        # add to styles list
        RadialProgressbar.styles.append(bar_style)

        # create style object
        s = ttk.Style()

        RadialProgressbar.imgs.setdefault(self.size, {})
        self.indicator_img = self.indicator_img or RadialProgressbar.imgs[self.size].get(self.fg)
        self.base_img = self.base_img or RadialProgressbar.imgs[self.size].get(self.bg)

        if not self.indicator_img:
            img = create_circle(self.size, color=self.fg)
            self.indicator_img = ImageTk.PhotoImage(img)
            RadialProgressbar.imgs[self.size].update(**{self.fg: self.indicator_img})

        if not self.base_img:
            img = create_circle(self.size, color=self.bg)
            self.base_img = ImageTk.PhotoImage(img)
            RadialProgressbar.imgs[self.size].update(**{self.bg: self.base_img})

        # create elements
        indicator_element = f'top_img_{bar_style}'
        base_element = f'bottom_img_{bar_style}'

        try:
            s.element_create(base_element, 'image', self.base_img, border=0, padding=0)
        except:
            pass

        try:
            s.element_create(indicator_element, 'image', self.indicator_img, border=0, padding=0)
        except:
            pass

        # create style layout
        s.layout(bar_style,
                 [(base_element, {'children':
                        [('pbar', {'side': 'left', 'sticky': 'nsew', 'children':
                                [(indicator_element, {'sticky': 'nswe'})]})]})])

        # configure new style
        s.configure(bar_style, pbarrelief='flat', borderwidth=0, troughrelief='flat')

        return bar_style

    def show_percentage(self, *args):
        """display progressbar percentage in a label"""
        bar_value = self.get()
        self.percent_label.config(text=f'{bar_value}%')

    def config(self, **kwargs):
        """config widgets' parameters"""

        # create style object
        s = ttk.Style()

        kwargs = {k: v for k, v in kwargs.items() if v}
        self.__dict__.update(kwargs)

        # frame bg
        self['bg'] = self.parent_bg

        # bar style configure
        s.configure(self.bar_style, background=self.parent_bg, troughcolor=self.parent_bg)

        # percentage label
        self.percent_label.config(background=self.text_bg, foreground=self.text_fg, font=self.font)


class RadialProgressbar3d(RadialProgressbar):
    """create radial 3d progressbar

        basically this is a ttk horizontal progressbar modified by using custom style layout and images

        Example:
            bar = Radial3dProgressbar(root, size=150, fg='blue')
            bar.pack(padx=10, pady=10)
            bar.start()
        """
    imgs = {}  # imgs{"size":{"color": img}}  example: imgs{"100":{"red": img}}

    def __init__(self, parent, size=100, fg='cyan', text_bg = '#333', text_fg = 'white', **extra):
        """initialize progressbar

        Args:
            parent  (tkinter object): tkinter container, i.e. toplevel window or frame
            size (int or 2-tuple(int, int)) size of progressbar
            fg(str): color of indicator ring
            extra: any extra kwargs, e.g. font, font_size_ratio, etc... see parent class docs
        """

        RadialProgressbar3d.imgs.setdefault(size, {})
        base_img = RadialProgressbar3d.imgs[size].get('base')
        indicator_img = RadialProgressbar3d.imgs[size].get(fg)

        if not indicator_img:
            # create pillow images
            base_img = create_pil_image(b64=progressbar_3d_base)
            indicator_img = create_circle(size=84, thickness=4, color=fg)

            # change indicator color
            indicator_img = change_img_color(indicator_img, fg)

            # merge indicator ring with base image copy
            indicator_img = mix_images(base_img.copy(), indicator_img)

            # resize
            if self.size:
                base_img = resize_img(base_img, size)
                indicator_img = resize_img(indicator_img, size)

            # create tkinter images using pillow ImageTk
            indicator_img = ImageTk.PhotoImage(indicator_img)
            base_img = ImageTk.PhotoImage(base_img)

            # store images for future use
            RadialProgressbar3d.imgs[size].update(**{'base': base_img})
            RadialProgressbar3d.imgs[size].update(**{fg: indicator_img})

        kwargs = locals().copy()
        kwargs.update(**extra)
        kwargs.pop('self')
        kwargs.pop('extra')

        RadialProgressbar.__init__(self, **kwargs) 
